import inspect
import sys
from .utils import Client
import threading
from threading import Thread
import asyncio
import atexit
import time
import uuid
import os
import ctypes
import logging
import aiohttp

debugger = threading.local()
debugger.instance = None

_logger = logging.getLogger(__name__)


def safe_serialize_dict(obj):
    res = {}
    for key, value in obj.items():
        if isinstance(value, (str, int, float, )):
            res[key] = value
        else:
            try:
                res[key] = repr(value)
            except Exception:
                res[key] = "Couldn't convert to string"
    return res


def safe_serialize_tuple(obj):
    if isinstance(obj, tuple):
        return [
            repr(val)
            for val in obj
        ]
    else:
        return repr(obj)


class DbClient(Client):
    def __init__(self, websocket, debugger):
        super(DbClient, self).__init__(websocket)
        self.debugger = debugger
        debugger.client_obj = self
        self.break_files = []

    async def on_read(self, data):
        if (
            data.get("action") == "configure" and
            data.get('type') == 'break'
        ):
            self.break_files = data.get('files')
            return

        if (
            data.get("action") == "interrupt"
        ):
            self.debugger.mode = 'step'

        if (
            data.get("action") == "stop"
        ):
            await self.stop()

        await super(DbClient, self).on_read(data)

    async def on_write(self, data):
        pass

    # async def event_loop(self):
    #     if self.debugger.stopped:
    #         self.on_exit()
    #         return
    #     await super(Client, self).event_loop()


class FakeDebugger(object):
    def __init__(self):
        self.stopped = True

    def stop(self):
        pass


def client_thread(parent_debugger):
    try:
        if debugger.instance:
            debugger.instance.stopped = True
    except Exception:
        debugger.instance = FakeDebugger()

    def get_tags():
        env = os.environ.copy()
        tags = []
        for key, value in env.items():
            if key.startswith('AWDB_TAGS_'):
                tags += [
                    val.strip()
                    for val in value.split(',')
                ]

        tags.append("thread:{}".format(threading.current_thread().getName()))

        return tags

    async def connect():
        HOSTNAME = os.environ.get('AWDB_URL', 'wss://awdb.docker/ws')
        VERIFY_SSL = os.environ.get('AWDB_VERIFY_SSL', 'TRUE')
        verify_ssl = VERIFY_SSL == "TRUE"
        connector = aiohttp.TCPConnector(verify_ssl=verify_ssl)
        session = aiohttp.ClientSession(connector=connector)
        websocket = await session.ws_connect(HOSTNAME)

        client = DbClient(websocket, parent_debugger)
        client.loop = asyncio.get_event_loop()
        client.uuid = uuid.uuid1()

        init_session = {
            "event": "new_session",
            "uuid": client.uuid.hex,
            "tags": get_tags()
        }

        await websocket.send_json(init_session)

        while True:
            if parent_debugger.stopped:
                await client.stop()
                return

            await client.event_loop()

        _logger.debug("Ended client")
        await session.close()

    loop = asyncio.new_event_loop()
    task = asyncio.ensure_future(connect(), loop=loop)
    loop.run_until_complete(task)
    parent_debugger.stopped = True


class Debugger(object):
    def __init__(self):
        self.client = Thread(target=client_thread, args=(self,), daemon=True)
        self.client.start()
        self.client_obj = None
        self.stopped = False

        self.thread_name = threading.current_thread().getName()
        self.breakpoints = []
        self.mode = 'step'
        self.break_on_return = False
        self.return_frame = None

        while not self.client_obj:
            _logger.debug(
                "[%s] Waiting for the client from the async thread ",
                self.thread_name
            )
            time.sleep(1)

    def test_breakpoints(self, frame, event):
        filename = frame.f_code.co_filename
        lineno = frame.f_lineno

        for b_event, b_lineno, b_filename in self.breakpoints:
            event_match = b_event == event or b_event == 'any'
            filename_match = b_filename == filename or b_filename == 'any'

            if b_lineno == 'any':
                lineno_match = True
            else:
                if isinstance(b_lineno, (tuple, list)):
                    lineno_match = b_lineno[0] <= lineno <= b_lineno[1]
                else:
                    lineno_match = b_lineno == lineno

            _logger.debug(
                "{} {} {}".format(
                    (event, b_event, event == b_event),
                    (lineno, b_lineno, lineno == b_lineno),
                    (filename, b_filename, filename == b_filename),
                )
            )

            if all([event_match, filename_match, lineno_match]):
                return True

    def should_break(self, frame, event, arg):
        if (
            self.break_on_return and
            event == 'return' and
            frame == self.return_frame
        ):
            self.break_on_return = False
            return True

        if self.test_breakpoints(frame, event):
            return True

        if self.mode == 'step':
            return True

        return False

    def inspect_frame(self, frame, event, arg):
        source_code, source_line = inspect.getsourcelines(frame.f_code)
        source = "".join(source_code)
        frame_locals = safe_serialize_dict(frame.f_locals)
        frame_globals = safe_serialize_dict(frame.f_globals)

        data = {
            "type": "frame",
            "session": self.client_obj.uuid.hex,
            "event": event,
            "arg": safe_serialize_tuple(arg),
            "source": source,
            "source_lineno": source_line,
            "lineno": frame.f_lineno,
            "filename": frame.f_code.co_filename,
            "locals": frame_locals,
            "globals": frame_globals
        }

        return data

    def __call__(self, frame, event, arg):
        if self.stopped or not self.client_obj:
            return

        if not self.should_break(frame, event, arg):
            return self

        thread_name = threading.current_thread().getName()
        _logger.debug(
            "Start Trace (%s) %s %s",
            thread_name,
            event,
            arg
        )

        data = self.inspect_frame(frame, event, arg)

        frame_stack = []
        current_frame = frame

        self.send_queue(data)

        while True:
            obj = self.receive_queue()

            if obj['action'] in ['step', 's']:
                self.mode = 'step'
                return self
            if obj['action'] in ['return', 'r']:
                self.break_on_return = True
                self.return_frame = frame
                self.mode = 'continue'
                return self
            elif obj['action'] == 'stop':
                _logger.debug("[%s] Removing trace to continue", thread_name)
                self.mode = 'step'
                sys.settrace(None)
                return
            elif obj['action'] in ['continue', 'c']:
                _logger.debug("[%s] Removing trace to continue", thread_name)
                self.mode = 'continue'
                return self
            elif obj.get('action') == 'set_locals':
                frame.f_locals.update({
                    obj.get('key'): obj.get('value')
                })
                ctypes.pythonapi.PyFrame_LocalsToFast(
                    ctypes.py_object(frame), ctypes.c_int(0)
                )
            elif obj.get('action') == 'execute':
                try:
                    code = compile(obj.get('code'), '<eval>', 'eval')
                    ret_val = eval(code, frame.f_globals, frame.f_locals,)
                except Exception:
                    local_values = frame.f_locals.copy()
                    try:
                        code = compile(obj.get('code'), '<exec>', 'exec')
                        ret_val = eval(code, frame.f_globals, local_values)
                        frame.f_locals.update(local_values)
                        ctypes.pythonapi.PyFrame_LocalsToFast(
                            ctypes.py_object(frame), ctypes.c_int(0)
                        )
                    except Exception:
                        _logger.error("Got error", exc_info=True)
                        ret_val = None

                self.send_queue({
                    "type": "retval",
                    "value": repr(ret_val)
                })
            elif obj.get('action') == 'add_breakpoint':
                _logger.debug("Adding breakpoint")
                self.breakpoints.append(obj.get('breakpoint'))
            elif obj.get('action') == 'remove_breakpoint':
                self.breakpoints.remove(obj.get('breakpoint'))
            elif obj.get('action') == 'breakpoints':
                self.send_queue({
                    "type": "breakpoints",
                    "breakpoints": list(self.breakpoints)
                })
            elif obj.get('action') == 'inspect':
                data = self.inspect_frame(current_frame, event, arg)
                self.send_queue(data)
            elif obj.get('action') == 'up':
                if (current_frame.f_back):
                    frame_stack.append(current_frame)
                    current_frame = current_frame.f_back
                data = self.inspect_frame(current_frame, event, arg)
                self.send_queue(data)
            elif obj.get('action') == 'down':
                if len(frame_stack) > 0:
                    current_frame = frame_stack.pop()
                data = self.inspect_frame(current_frame, event, arg)
                self.send_queue(data)
            elif obj.get('action') == 'w':
                action = self.get_call_stack(current_frame)
                self.send_queue(action)
            else:
                _logger.info(
                    "Unknown command waiting for next command"
                )

            self.send_queue({'action': 'none'})

        return self

    def get_call_stack(self, frame):
        frames = inspect.getouterframes(frame)

        frames.reverse()

        return {
            "type": "frames",
            "frames": [
                {
                    "filename": info.filename,
                    "lineno": info.lineno,
                    "function": info.function
                }
                for info in frames
            ]
        }

    def send_queue(self, data):
        queue = self.client_obj.send_queue

        async def put_queue(data):
            _logger.debug("Put data into queue")
            return await queue.put(data)

        loop = self.client_obj.loop
        asyncio.run_coroutine_threadsafe(put_queue(data), loop).result()

    def receive_queue(self):
        async def get_queue():
            return await self.client_obj.recv_queue.get()

        loop = self.client_obj.loop
        return asyncio.run_coroutine_threadsafe(get_queue(), loop).result()

    def stop(self):
        self.stopped = True

        if sys.gettrace() == self:
            sys.settrace(None)

        if self.client_obj:
            self.client_obj.stopped = True

        self.client.join()


def set_trace():
    _logger.debug("[%s] Set trace", threading.current_thread().getName())
    thread_name = threading.current_thread().getName()

    try:
        instance = debugger.instance
        if not instance or instance.stopped:
            debugger.instance = Debugger()
            instance = debugger.instance
    except Exception:
        debugger.instance = Debugger()
        instance = debugger.instance

    def destroy(instance):
        def wrap():
            _logger.debug("[%s] Destroying debugger instance", thread_name)
            instance.stop()
        return wrap

    atexit.register(destroy(instance))

    frame = sys._getframe().f_back
    while frame:
        frame.f_trace = instance
        frame = frame.f_back

    if not instance.stopped:
        _logger.debug("[%s] Set trace to %s", thread_name, instance)
        if thread_name == 'MainThread':
            sys.settrace(instance)
        else:
            sys.settrace(instance)
    else:
        _logger.debug("[%s] Remove trace", thread_name)
        if thread_name == 'MainThread':
            sys.settrace(None)
        else:
            sys.settrace(None)



# -*- coding: utf-8 -*-
import asyncio
# import janus
import json
import logging

_logger = logging.getLogger(__name__)


class Read(object):
    def __init__(self, value):
        self.value = value


class Write(object):
    def __init__(self, value):
        self.value = value


class Client(object):
    def __init__(self, socket):
        self.socket = socket
        self.send_queue = asyncio.Queue()
        self.recv_queue = asyncio.Queue()
        self.read_task = None
        self.write_task = None
        self.timeout = 1
        self.message = None
        self.events = []
        self.stopped = False

    async def on_read(self, data):
        await self.recv_queue.put(data)

    async def on_write(self, data):
        pass

    async def event_loop(self):
        if self.stopped:
            return await self.stop()

        async def write():
            if not self.message:
                self.message = await self.send_queue.get()

            message = self.message
            _logger.debug("Sending message [%s]", self.message)
            await self.socket.send_json(self.message)
            _logger.debug("Message sent")
            self.message = None
            return message

        async def read():
            data = await self.socket.receive_json()
            _logger.debug("Read data %s", data)
            return data

        if not self.read_task:
            self.read_task = asyncio.ensure_future(read())

        if not self.write_task:
            self.write_task = asyncio.ensure_future(write())

        done, pending = await asyncio.wait(
            {self.read_task, self.write_task},
            timeout=self.timeout,
            return_when=asyncio.FIRST_COMPLETED
        )

        _logger.debug("Done Jobs %s %s", done, self.uuid)

        if self.read_task in done:
            try:
                self.events.append(
                    asyncio.ensure_future(
                        self.on_read(await self.read_task)
                    )
                )
            except Exception:
                self.stopped = True
            finally:
                self.read_task = None

        if self.write_task in done:
            self.events.append(
                asyncio.ensure_future(
                    self.on_write(await self.write_task)
                )
            )
            self.write_task = None

        if self.events:
            done, pending = await asyncio.wait(self.events, timeout=0.1)
            for task in done:
                await task
                self.events.remove(task)

    async def stop(self):
        self.stopped = True
        await self.on_exit()

    async def on_exit(self):
        if self.read_task:
            self.read_task.cancel()
            self.read_task = None

        if self.write_task:
            self.write_task.cancel()
            self.write_task = None

        for task in self.events:
            task.cancel()

        self.events = []

        try:
            await self.socket.close()
        except Exception:
            _logger.error("Error closing socket")



import asyncio
from aiohttp import web
from ..utils import Client
import logging

_logger = logging.getLogger(__name__)


class DebuggerClient(Client):

    def __init__(self, websocket):
        super(DebuggerClient, self).__init__(websocket)
        self.subscribers = set()

    def subscribe(self, client):
        self.subscribers.add(client)

    def unsubscribe(self, client):
        self.subscribers.remove(client)

    async def on_read(self, dat):
        for client in self.subscribers:
            await client.send_queue.put(dat)


class DebuggingClient(Client):
    def __init__(self, websocket, servers):
        super(DebuggingClient, self).__init__(websocket)
        self.servers = servers

    async def on_read(self, data):
        if data.get('action') == 'list':
            await self.send_queue.put({
                "type": "instances",
                "instances": [
                    {
                        "uuid": server.uuid,
                        "tags": server.tags
                    }
                    for server in
                    self.servers
                ]
            })
        elif data.get('action') == 'subscribe':
            for server in self.servers:
                if server.uuid == data.get('uuid'):
                    server.subscribe(self)
        elif data.get('action') == 'unsubscribe':
            for server in self.servers:
                if server.uuid == data.get('uuid'):
                    server.unsubscribe(self)
        elif data.get('action') == 'call':
            for server in self.servers:
                if server.uuid == data.get('uuid'):
                    await server.send_queue.put(data.get('params'))
        elif data.get('action') == 'add_user':
            username = data.get('username')
            if username not in users:
                users[username] = set()
        elif data.get('action') == 'rm_user':
            username = data.get('username')
            if username in users:
                del users[username]
        elif data.get('action') == 'kill':
            to_remove = set()

            for server in self.servers:
                if server.uuid == data.get('uuid'):
                    to_remove.add(server)

            for server in to_remove:
                self.servers.remove(server)
                await server.close()

debugging_sessions = set()
debugging_clients = set()
users = dict()


async def start_aiohttp():
    import os

    logging.basicConfig(level='DEBUG')

    _logger.info("Starting server")

    async def websocket_handler(request):
        _logger.info("New Request")
        done = False
        client = None
        ws = web.WebSocketResponse()
        await ws.prepare(request)

        _logger.info("New connexion")

        data = await ws.receive_json()
        print(data)

        if data.get('event') == 'new_session':
            _logger.info("Got new debug")
            client = DebuggerClient(ws)
            client.uuid = data.get('uuid')
            client.tags = data.get('tags', [])
            debugging_sessions.add(client)
        elif data.get('event') == 'new_client':
            _logger.info("Got new client")
            data = await ws.receive_json()

            if data.get('action') == 'authenticate':
                if data.get('token') not in users:
                    done = True
                else:
                    client = DebuggingClient(ws, debugging_sessions)
                    client.uuid = data.get('client_id')
                    client.token = data.get('token')
                    client.user_info = data.get('client_info')

                    users[client.token].add(client)
                    debugging_clients.add(client)
            else:
                done = True

        while not done:
            await client.event_loop()
            try:
                if client.stopped:
                    await client.stop()
                    break
            except KeyboardInterrupt:
                await client.on_read(None)

        if isinstance(client, DebuggerClient):
            debugging_sessions.remove(client)
        elif isinstance(client, DebuggingClient):
            debugging_clients.remove(client)
            users[client.token].remove(client)

        _logger.info('websocket connection closed')

        return ws

    if os.environ.get('AWDB_ADMIN_TOKEN'):
        users[os.environ.get('AWDB_ADMIN_TOKEN')] = set()

    app = web.Application()
    bind_host = os.environ.get('AWDB_BIND_HOST', 'localhost')
    bind_port = int(os.environ.get('AWDB_BIND_PORT', '8080'))
    app.add_routes([web.get('/ws', websocket_handler)])

    runner = web.AppRunner(app)
    await runner.setup()
    site = web.TCPSite(runner, bind_host, bind_port)
    _logger.info("Serving on http://{}:{}".format(bind_host, bind_port))
    await site.start()

    while True:
        try:
            await asyncio.sleep(3600)
        except KeyboardInterrupt:
            break

    await runner.cleanup()


def server():
    logging.basicConfig(level='INFO')
    loop = asyncio.get_event_loop()
    app = start_aiohttp()
    loop.run_until_complete(app)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    requirements = fh.read()

setuptools.setup(
    name="awdb",
    version="0.0.9",
    author="Archeti",
    author_email="info@archeti.ca",
    description="WebSocket Debugger",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/archeti-org/awdb.git",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    entry_points={
        "console_scripts": [
            "awdb-server = awdb.cli:server",
            "awdb-client = awdb.cli:client",
        ],
    },
    # include_package_data=True,
    package_data={
        "awdb": [
            "static/*",
            "static/**/*"
        ],
    }
)



import ast
import json
import logging
import os
import tempfile
from subprocess import call

_logger = logging.getLogger(__name__)

logging.basicConfig(level='INFO')


class JSON(object):
    def __init__(self, data):
        self.data = data

    def __format__(self, format):
        return json.dumps(self.data, indent=2, sort_keys=True)


frame_template = """
{session}
{event}
{filename}:{lineno}:{source_lineno}
{arg_json}

Locals:
{locals_json}

Globals:
{globals_json}

Code:
{source}
"""


def parse_break_line(command):
    _, rest = command.split(' ', 1)
    event, rest = rest.split(' ', 1)
    line, filename = rest.split(' ', 1)

    if ':' in line:
        line = line.split(':', 1)
        line[0] = int(line[0])
        line[1] = int(line[1])
    else:
        line = int(line) if line != 'any' else line

    return [event, line, filename]


def parse_command(command):
    if command in [
        'step',
        's',
        'continue',
        'c',
        'stop',
        'inspect',
        'return',
        'interrupt',
        'breakpoints',
        'w',
        'up',
        'down',
    ]:
        return {'action': command}
    elif command.startswith('set'):
        _, vals = command.split(' ', 1)
        key, value = command.split(' ')
        action = {
            'action': 'set_locals',
            'key': key,
            'value': ast.literal_eval(value)
        }
        return action
    elif command.startswith('eval+'):
        editor = os.environ.get('editor', 'vim')
        with tempfile.NamedTemporaryFile(suffix='.py') as fin:
            call([editor, fin.name])
            fin.seek(0)
            code = fin.read().decode('utf-8')
        print(code)
        action = {
            'action': 'execute',
            'code': code
        }
        return action
    elif command.startswith('eval'):
        _, code = command.split(' ', 1)
        action = {
            'action': 'execute',
            'code': code
        }
        return action
    elif command.startswith('break'):
        action = {
            "action": "add_breakpoint",
            "breakpoint": parse_break_line(command)
        }
        return action
    elif command.startswith('-break'):
        action = {
            "action": "remove_breakpoint",
            "breakpoint": parse_break_line(command)
        }
        return action
    else:
        action = {
            'action': 'execute',
            'code': command
        }
        return action


def format_response(data):
    if data.get('type') == 'frame':
        lineno = data.get('lineno', 0) - data.get('source_lineno', 0)

        source = "\n".join([
            "{}{}".format(
                '->' if index == lineno else '  ',
                line
            )
            for index, line in enumerate(data.get('source').split('\n'))
        ])

        data['source'] = source

        data.update({
            "locals_json": JSON(data['locals']),
            "globals_json": JSON(data['globals']),
            "arg_json": JSON(data['arg']),
        })

        print(frame_template.format(**data))
    elif data.get('type') == 'frames':
        print("Call stack")
        for info in data.get('frames', []):
            print(
                " {filename}:{lineno} in {function}".format(**info)
            )
    elif data.get('type') == 'instances':
        print("Instances")
        for instance in data.get('instances', []):
            print("{uuid}: {tags}".format(**instance))
    elif data.get('type') == 'breakpoints':
        print("Breakpoints")
        for event, line, filename in data.get('breakpoints', []):
            print("Event: {} File: {}:{}".format(event, filename, line))
    elif data.get('type') == 'retval':
        print(data.get('value'))


def client():
    import aiohttp
    import asyncio
    import os

    HOSTNAME = os.environ.get('AWDB_URL', 'wss://awdb.docker/ws')
    print(HOSTNAME)

    async def get_command(active):
        from aioconsole import ainput
        prompt = "({})".format(active) if active else ""
        return await ainput("command {}> ".format(prompt))

    async def prompt(prompt):
        from aioconsole import ainput
        return await ainput("{}> ".format(prompt))

    async def execute_commands():
        connector = aiohttp.TCPConnector(verify_ssl=False)
        session = aiohttp.ClientSession(connector=connector)
        websocket = await session.ws_connect(HOSTNAME)

        active = None
        read = None

        await websocket.send_json({
            "event": "new_client"
        })

        token = await prompt("Token")
        username = await prompt("Username")

        await websocket.send_json({
            "action": "authenticate",
            "token": token,
            "client_id": 1,
            "client_info": {
                "username": username,
            }
        })

        while True:
            input_task = asyncio.ensure_future(get_command(active))
            if not read:
                read = asyncio.ensure_future(websocket.receive_json())

            done, pending = await asyncio.wait(
                [input_task, read],
                return_when=asyncio.FIRST_COMPLETED
            )

            if read in done:
                response = await read
                format_response(response)
                read = None

            if input_task not in done:
                input_task.cancel()
                continue

            command = await input_task

            # prompt = "({})".format(active) if active else ""
            # command = input("command {}> ".format(prompt))

            action = None
            if not command:
                continue
                # if read:
                #     read.cancel()
                # break

            if command == 'list':
                action = {
                    'action': 'list'
                }
            elif command.startswith('use'):
                active = command.split(' ', 1)[1]
            elif command.startswith('subscribe'):
                active = command.split(' ', 1)[1]
                action = {
                    'action': 'subscribe',
                    'uuid': active
                }
            elif command.startswith('unsubscribe'):
                active = command.split(' ', 1)[1]
                action = {
                    'action': 'unsubscribe',
                    'uuid': active
                }
            elif command.startswith('add_user'):
                user = command.split(' ', 1)[1]
                action = {
                    'action': 'add_user',
                    'username': user
                }
            elif command.startswith('rm_user'):
                user = command.split(' ', 1)[1]
                action = {
                    'action': 'rm_user',
                    'username': user
                }
            elif command.startswith('kill'):
                action = {
                    'action': 'kill',
                    'uuid': active,
                }
                active = None
            else:
                action = parse_command(command)
                if action:
                    action = {
                        'action': 'call',
                        'params': action,
                        'uuid': active
                    }

            if action:
                await websocket.send_json(action)

                # while True:
                #     read = asyncio.create_task(websocket.receive_json())
                #     done, pending = await asyncio.wait({read}, timeout=1)
                #     if len(done) > 0:
                #         for dd in done:
                #             response = await dd
                #             format_response(response)
                #     else:
                #         read.cancel()
                #         break

        await websocket.close()
        await session.close()

        print("done")

    loop = asyncio.get_event_loop()
    task = asyncio.ensure_future(execute_commands())
    loop.run_until_complete(task)
